# This needs to sync with MANIFEST.in

exclude_website_pats = [
    "**/.DS_Store",
    "examples/**",
    "_posts/**",
    "HISTORY.md",
    "docs/benchmarks*md",
    "docs/ctags.md",
    "docs/unified-diffs.md",
    "docs/leaderboards/index.md",
    "assets/**",
    ".jekyll-metadata",
    "Gemfile.lock",
    "Gemfile",
    "_config.yml",
    "**/OLD/**",
    "OLD/**",
]
